"""Display job progress for FFmpeg audio extraction."""

from bidict import MutableBidirectionalMapping, bidict
from rich.console import Console, ConsoleOptions, ConsoleRenderable, Group, RenderResult
from rich.highlighter import ReprHighlighter
from rich.padding import Padding
from rich.progress import BarColumn, Progress, TaskID, TextColumn, TimeElapsedColumn, TimeRemainingColumn

import ffmpeg
from BAET._config.console import app_console
from BAET._config.logging import create_logger
from BAET.FFmpeg.jobs import AudioExtractJob, stream_duration_ms
from BAET.typing import StreamTaskBiMap

logger = create_logger()


class FFmpegJobProgress(ConsoleRenderable):
    """Job progress display for FFmpeg audio extraction.

    Attributes
    ----------
    job : AudioExtractJob
    """

    # TODO: Need mediator to consumer/producer printing
    def __init__(self, job: AudioExtractJob) -> None:
        self.job = job

        bar_blue = "#5079AF"
        bar_yellow = "#CAAF39"

        self._overall_progress = Progress(
            TextColumn('Progress for "{task.fields[filename]}"', highlighter=ReprHighlighter()),
            BarColumn(
                complete_style=bar_blue,
                finished_style="green",
                pulse_style=bar_yellow,
            ),
            TextColumn("Completed {task.completed} of {task.total}", highlighter=ReprHighlighter()),
            TimeElapsedColumn(),
            TimeRemainingColumn(),
            console=app_console,
        )

        self._overall_progress_task = self._overall_progress.add_task(
            "Waiting...",
            start=False,
            filename=job.input_file.name,
            total=len(self.job.audio_streams),
        )

        self._stream_task_progress = Progress(
            TextColumn("Audio stream {task.fields[stream_index]}", highlighter=ReprHighlighter()),
            BarColumn(
                complete_style=bar_blue,
                finished_style="green",
                pulse_style=bar_yellow,
            ),
            TextColumn("{task.fields[status]}"),
            TimeElapsedColumn(),
            TimeRemainingColumn(),
            console=app_console,
        )

        self._stream_task_bimap: StreamTaskBiMap = bidict()

        stream_task_bimap: MutableBidirectionalMapping[int, TaskID] = bidict()
        for stream in self.job.audio_streams:
            stream_index = stream["index"]

            task = self._stream_task_progress.add_task(
                "Waiting...",
                start=False,
                total=stream_duration_ms(stream),
                stream_index=stream_index,
                status="[plum4]Waiting[/]",
            )

            stream_task_bimap[stream_index] = task

        self._stream_task_bimap = stream_task_bimap

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        """Render the job progress display.

        Parameters
        ----------
        console : Console
            The console to render to.
        options : ConsoleOptions
            The console options.

        Returns
        -------
        RenderResult
            The render result.
        """
        yield Group(
            self._overall_progress,
            Padding(self._stream_task_progress, (1, 0, 1, 5)),
        )

    def _run_task(self, task: TaskID) -> None:
        stream_index: int = self._stream_task_bimap.inverse[task]

        logger.info("Extracting audio stream %d of %r", stream_index, self.job.input_file.name)

        output = self.job.stream_indexed_outputs[stream_index]

        logger.debug("Running: %s", " ".join(ffmpeg.compile(output)))

        proc = ffmpeg.run_async(
            output,
            pipe_stdout=True,
            pipe_stderr=True,
        )

        try:
            with proc as p:
                if p is None:
                    raise ValueError("FFmpeg process failed to start")

                if p.stdout is None:
                    raise ValueError("FFmpeg process stdout is None")

                for line in p.stdout:
                    decoded = line.decode("utf-8").strip()
                    if "out_time_ms" in decoded:
                        val = decoded.split("=", 1)[1]
                        self._stream_task_progress.update(
                            task,
                            completed=float(val),
                        )

                err = p.stderr.read().strip() if p.stderr is not None else b"No stderr output was captured."
            if proc.wait() != 0:
                raise RuntimeError(err.decode("utf-8"))
        except (RuntimeError, ValueError) as e:
            logger.critical("%s: %s", type(e).__name__, e)
            raise e

    def start(self) -> None:
        """Start the job and render the progress display."""
        self._overall_progress.start_task(self._overall_progress_task)
        logger.info("Stream index to job task ID bimap: %r", self._stream_task_bimap)
        for task in self._stream_task_bimap.values():
            self._stream_task_progress.start_task(task)

            self._stream_task_progress.update(task, status="[italic cornflower_blue]Working[/]")

            try:
                self._run_task(task)
                self._stream_task_progress.update(task, completed=self._stream_task_progress.tasks[task].total)
                self._stream_task_progress.update(task, status="[bold green]Complete[/]")
            except (RuntimeError, ValueError):
                self._stream_task_progress.update(task, status="[bold red]ERROR[/]")
            finally:
                self._stream_task_progress.stop_task(task)
                self._overall_progress.advance(self._overall_progress_task, advance=1)

        self._overall_progress.stop_task(self._overall_progress_task)



"""Helper methods for string manipulation."""

from collections.abc import Callable, Sequence


def pretty_join[T](
    items: Sequence[T],
    message: str,
    *,
    bullet: str = "∙ ",
    formatter: Callable[[T], str] | None = None,
    force_newline: bool = False,
) -> str:
    """Join a sequence of items into a pretty list string.

    Parameters
    ----------
    items : Sequence[T]
        A sequence of strings to join.

    message : str
        The message to prepend to the list.

    bullet : str
        The bullet to use for each item.

    formatter : Callable[[T], str], optional
        A function to format each item. If None, then no formatter will be used.

    force_newline : bool
        If True, then a newline will be forced even if there is only one item.


    Returns
    -------
    str
        The prettified list string.


    Examples
    --------
    >>> items = ["one", "two", "three", "four"]
    >>> pretty_join(items, "Test items")
    """

    def _formatter(item: T) -> str:
        return f"'{item}'"

    if formatter is None:
        formatter = _formatter

    if not force_newline and len(items) == 0:
        return f"{message}: []"
    elif not force_newline and len(items) == 1:
        return f"{message}: {formatter(items[0])}"

    newline_base = f"\n{' ' * 4}{bullet}"
    joiner = f",{newline_base}"
    formatted = joiner.join([formatter(p) for p in items])
    return f"{message}:{newline_base}{formatted}"


if __name__ == "__main__":
    items = ["one", "two", "three", "four"]
    print(pretty_join(items, "Test items"))
    print(pretty_join(["world"], "hello", formatter=lambda x: f"~{x}~"))



"""Typing definitions for BAET."""

from collections.abc import Mapping
from typing import Any

from bidict import BidirectionalMapping
from rich.progress import TaskID

from ffmpeg import Stream

# Numbers
type Millisecond = int | float

# FFmpeg
type StreamIndex = int
type AudioStream = dict[str, Any]
type FFmpegOutput = Stream

# Mappings
type IndexedOutputs = Mapping[StreamIndex, FFmpegOutput]
type IndexedAudioStream = Mapping[StreamIndex, AudioStream]
type StreamTaskBiMap = BidirectionalMapping[StreamIndex, TaskID]



"""Context objects for commands with Click."""

from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from re import Pattern

import click


@dataclass()
class InOutArgs:
    """Argument for a command."""

    inputs: Sequence[Path]
    output: Path


@dataclass()
class InFilterArgs:
    """Argument for a command."""

    include_pattern: Pattern[str]
    exclude_pattern: Pattern[str] | None


@dataclass()
class Output:
    """Argument for a command."""

    verbose: bool
    dry_run: bool


@dataclass()
class CliOptions:
    """Argument for a command."""

    logging: bool = False
    dry_run: bool = False


pass_cli_options = click.make_pass_decorator(CliOptions, ensure=True)



import inspect
import logging
from collections.abc import Callable
from functools import wraps
from logging import FileHandler, Logger
from pathlib import Path
from types import ModuleType
from typing import Any, Concatenate

from rich.logging import RichHandler

from .console import app_console

rich_handler = RichHandler(
    rich_tracebacks=True,
    tracebacks_show_locals=True,
    console=app_console,
)

logging.basicConfig(
    level=logging.INFO,
    format="%(threadName)s: %(message)s",
    datefmt="[%X]",
    handlers=[rich_handler],
)

app_logger = logging.getLogger("app_logger")


def pass_module[**P](func: Callable[Concatenate[ModuleType, P], Logger]) -> Callable[P, Logger]:
    """Decorate a function, passing the module of the caller as the first argument of the decorated function."""
    frame = inspect.stack()[1]
    module = inspect.getmodule(frame[0])

    if module is None:
        raise RuntimeError("Could not inspect module")

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Any:
        return func(module, *args, **kwargs)

    return wrapper


@pass_module
def create_logger(module: ModuleType) -> Logger:
    """Create and return a logger for the current module."""
    if module is None:
        raise RuntimeError("Could not inspect module")

    module_name = module.__name__

    logger = app_logger.getChild(module_name)
    return logger


@pass_module
def find_module_logger(module: ModuleType) -> Logger:
    """Find the logger for the current module.

    Parameters
    ----------
    module : ModuleType
        The module to find the logger for.

    Returns
    -------
    Logger
        The logger for the module, or None if it doesn't exist.

    Raises
    ------
    TypeError
        If the found logger variable is not of type logging.Logger.
    """
    logger = getattr(module, "logger", app_logger)

    if not isinstance(logger, logging.Logger):
        raise TypeError("logger must be of type logging.Logger")

    return logger


def pass_logger(func: Callable[..., Any]) -> Callable[..., Any]:
    """Pass the modules logger to the decorated function, if it exists. Otherwise, uses the root logger.

    Parameters
    ----------
    func : Callable[..., Any]
        The function to decorate.

    Returns
    -------
    Callable[..., Any]
        The newly decorated function.

    Raises
    ------
    RuntimeError
        If the module cannot be inspected.
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        logger = find_module_logger()

        if logger is None:
            logger = app_logger

        return func(logger, *args, **kwargs)

    return wrapper


def configure_logging(*, enable_logging: bool = True, file_out: Path | None = None) -> None:
    """Configure logging.

    Parameters
    ----------
    enable_logging : bool, optional
        Whether to enable logging, by default True
    file_out : Path | None, optional
        The file to write logs to, by default None
    """
    if not enable_logging:
        logging.disable(logging.INFO)

    if file_out is not None:
        handler = FileHandler(filename=file_out)
        handler.setFormatter(rich_handler.formatter)
        app_logger.addHandler(handler)



"""Conversion utility methods."""

import datetime


def micro_to_hhmmss(micro: float) -> str:
    """Convert milliseconds to HH:MM:SS format."""
    return str(datetime.timedelta(microseconds=micro))


if __name__ == "__main__":
    ms = 4007829000.0
    print(micro_to_hhmmss(ms))



"""Jobs that encapsulate work to be done by FFmpeg."""

import re
from collections.abc import Sequence
from fractions import Fraction
from logging import Logger
from pathlib import Path
from typing import Any, overload

import rich.repr
from more_itertools import first_true

import ffmpeg
from BAET._config.logging import create_logger
from BAET.cli.types import FFmpegArgsRepr
from BAET.helpers.string_helpers import pretty_join
from BAET.helpers.time_conversion import micro_to_hhmmss
from BAET.typing import AudioStream, FFmpegOutput, IndexedAudioStream, IndexedOutputs, Millisecond, StreamIndex

logger: Logger = create_logger()


@rich.repr.auto()
class JobMetadata:
    """Metadata for an FFmpeg job."""

    @overload
    def __init__(
        self,
        input_file: Path,
        output_dir: Path,
        *,
        audio_stream_indexes: Sequence[StreamIndex],
        job_outputs: Sequence[FFmpegOutput],
    ) -> None: ...

    @overload
    def __init__(
        self,
        input_file: Path,
        output_dir: Path,
        *,
        stream_indexed_output_map: IndexedOutputs,
    ) -> None: ...

    def __init__(
        self,
        input_file: Path,
        output_dir: Path,
        *,
        audio_stream_indexes: Sequence[StreamIndex] | None = None,
        job_outputs: Sequence[FFmpegOutput] | None = None,
        stream_indexed_output_map: IndexedOutputs | None = None,
    ):
        """Metadata for an FFmpeg job.

        Parameters
        ----------
        input_file : Path
            The input file path
        output_dir : Path
            The output directory path
        audio_stream_indexes : Sequence[StreamIndex] | None, optional
            The audio stream indexes to extract, by default None
        job_outputs : Sequence[FFmpegOutput] | None, optional
            The FFmpeg outputs for the job, by default None
        stream_indexed_output_map : IndexedOutputs | None, optional
            The map of stream indexes to FFmpeg outputs, by default None

        Raises
        ------
        ValueError
            Either `audio_stream_indexes` and `job_outputs` or `stream_indexed_output_map` must be provided

        Notes
        -----
        - If `audio_stream_indexes` and `job_outputs` are provided, they will be used to initialize the job metadata.
        - If `stream_indexed_output_map` is provided, it will be used to initialize the job metadata.
        - If both `audio_stream_indexes` and `stream_indexed_output_map` are provided,
            `audio_stream_indexes` will take precedence.
        """
        self.input_file = input_file
        self.output_dir = output_dir

        self.audio_stream_indexes = audio_stream_indexes
        self.job_outputs = job_outputs

        self.stream_indexed_output_map = stream_indexed_output_map

        if self.stream_indexed_output_map is not None:
            self.audio_stream_indexes, self.job_outputs = zip(
                *self.stream_indexed_output_map.items(),
                strict=True,
            )
            return

        if self.audio_stream_indexes is None or self.job_outputs is None:
            raise ValueError(
                "Either `audio_stream_indexes` and `job_outputs` or `stream_indexed_output_map` must be provided"
            )

        self.stream_indexed_output_map = dict(
            zip(
                self.audio_stream_indexes,
                self.job_outputs,
                strict=True,
            )
        )


def stream_duration_ms(stream: AudioStream) -> Millisecond:
    """Get the duration of the audio stream in milliseconds.

    Parameters
    ----------
    stream : AudioStream
        The audio stream to get the duration for

    Returns
    -------
    Millisecond
        The duration in milliseconds
    """
    if "duration_ts" in stream:
        # Convert the duration from seconds to microseconds
        return 1_000_000 * float(stream["duration_ts"]) * float(Fraction(stream["time_base"]))
    elif "tags" in stream:
        _, duration = first_true(
            stream["tags"].items(),
            pred=lambda x: x[0].lower() == "duration",
            default=(None, None),
        )

        if duration is None:
            raise ValueError(f"Could not find duration in {stream!r}")

        # Get the duration via regex. .strptime() has leftover data
        pattern = re.compile(r"(?P<H>\d+):(?P<M>\d+):(?P<S>\d+(?:\.\d+)?)")
        match = pattern.match(duration)
        if match:
            h, m, s = match.group("H", "M", "S")
            return 1_000_000 * (float(h) * 3600 + float(m) * 60 + float(s))
        else:
            raise ValueError(f"Could not parse duration from {duration!r}")

    raise ValueError(f"Could not find duration in {stream!r}")


class AudioExtractJob:
    """An FFmpeg job to extract audio from a video file.

    Attributes
    ----------
    input_file : Path
    stream_indexed_outputs : IndexedOutputs
    audio_streams : Sequence[AudioStream]
    indexed_audio_streams : IndexedAudioStream
    durations_ms_dict : dict[StreamIndex, Millisecond]
    """

    def __init__(
        self,
        input_file: Path,
        audio_streams: Sequence[AudioStream],
        indexed_outputs: IndexedOutputs,
    ) -> None:
        self.input_file: Path = input_file
        self.stream_indexed_outputs: IndexedOutputs = indexed_outputs
        self.audio_streams = audio_streams

        indexed_audio_streams = {}
        for stream in audio_streams:
            indexed_audio_streams[stream["index"]] = stream
        self.indexed_audio_streams: IndexedAudioStream = indexed_audio_streams

        # TODO: Do we need this?
        self.durations_ms_dict: dict[StreamIndex, Millisecond] = {
            stream["index"]: stream_duration_ms(stream) for stream in audio_streams
        }

        logger.info(
            pretty_join(
                list(self.durations_ms_dict.items()),
                "Parsed duration for %r",
                formatter=lambda kvp: f"Stream index {kvp[0]}: {kvp[1]} ({micro_to_hhmmss(kvp[1])})",
                force_newline=True,
            ),
            self.input_file,
        )

    def __rich_repr__(self) -> rich.repr.Result:
        """Return a rich representation of the object.

        Returns
        -------
        rich.repr.Result
            The rich representation.
        """
        yield "input_file", self.input_file
        yield "indexed_audio_streams", self.indexed_audio_streams
        yield "durations_ms_dict", self.durations_ms_dict
        yield (
            "stream_indexed_outputs",
            {k: FFmpegArgsRepr(ffmpeg.get_args(v)) for k, v in self.stream_indexed_outputs.items()},
        )

    def stream(self, index: StreamIndex) -> AudioStream:
        """Get the audio stream with the given index.

        Parameters
        ----------
        index : StreamIndex
            The index of the audio stream to get

        Returns
        -------
        AudioStream
            The audio stream with the given index
        """
        stream: AudioStream | None = first_true(
            self.audio_streams,
            default=None,
            pred=lambda st: st["index"] == index if st is not None else None,
        )

        if stream is None:
            raise IndexError(f'Stream with index "{index}" not found')

        return stream


if __name__ == "__main__":
    mapping: dict[StreamIndex, Any] = {x: f"Job {x:02d}" for x in range(1, 5)}

    meta = JobMetadata(
        input_file=Path("in"),
        output_dir=Path("out"),
        stream_indexed_output_map=mapping,
    )

    rich.print(meta)



"""Call FFprobe on a video file."""

from collections import ChainMap, OrderedDict
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import rich
import rich_click as click

from BAET._config.logging import create_logger
from BAET.cli.help_configuration import baet_config
from BAET.FFmpeg.probe import probe_file

logger = create_logger()

type _key_selector = Callable[[dict[str, Any]], dict[str, Any]]


@dataclass()
class ProbeContext:
    """Context for the probe command."""

    file: Path
    streams_only: bool
    tracks: tuple[int, ...] | None
    key: tuple[str, ...] | None


def _probe_file(file: Path) -> dict[str, Any]:
    probed_dict: OrderedDict[str, Any]
    with probe_file(file) as probed:
        if "format" in probed:
            probed_dict = OrderedDict(probed)
            probed_dict.move_to_end("format", last=False)

        return dict(probed)


@click.group(chain=True, invoke_without_command=True)
@baet_config(use_markdown=True)
@click.argument("file", type=click.Path(exists=True, dir_okay=False), required=True)
def probe(file: Path) -> None:
    """Call FFprobe on a video file."""


@probe.result_callback()
def probe_result_callback(commands: list[_key_selector], file: Path) -> None:
    """Run the probe command."""
    probed: dict[str, Any] = _probe_file(file)

    if not commands:
        rich.print_json(data=probed)
        return

    filtered: dict[str, Any] = dict(ChainMap(*[command(probed) for command in commands]))

    rich.print_json(data=filtered)


@probe.command(name="filter")
@click.option(
    "-a",
    "--audio",
    multiple=True,
    help="Specific audio track index to probe. Can be specified multiple times.",
    type=click.IntRange(min=0),
)
@click.option(
    "--format/--no-format",
    "-f/-nf",
    "format_",
    is_flag=True,
    default=None,
    show_default="Disabled if --audio is specified.",
    help="Include/Exclude format metadata.",
)
def probe_filter(audio: tuple[int, ...], format_: bool | None) -> _key_selector:
    """Specify which audio streams to probe."""
    if audio and format_ is None:
        format_ = False
    elif not audio and format_ is None:
        format_ = True

    logger.info("Filtering streams with index: %r", audio)
    logger.info("Including format information: %s", format_)

    def processor(meta: dict[str, Any]) -> Any:
        if "streams" not in meta:
            raise click.ClickException("No stream metadata found")

        format_dict = {"format": meta["format"]} if format_ else {}
        return format_dict | {"streams": _filter_streams(audio, meta["streams"])}

    return processor


def _filter_streams(indexes: Sequence[int], streams: list[dict[str, Any]]) -> list[dict[str, Any]]:
    if not indexes:
        return streams

    return [stream for stream in streams if stream["index"] in indexes]



"""Utilities for FFmpeg probe operations."""

import contextlib
from collections.abc import Iterator
from pathlib import Path
from typing import Any

import rich_click as click

import ffmpeg
from BAET._config.console import error_console
from BAET._config.logging import create_logger
from BAET.typing import AudioStream

logger = create_logger()


@contextlib.contextmanager
def probe_file(file: Path) -> Iterator[dict[str, Any]]:
    """Probe a file using FFmpeg."""
    logger.info("Probing file %r", file)

    try:
        probed: dict[str, Any] = ffmpeg.probe(file)
    except ffmpeg.Error as e:
        err: str = e.stderr.decode()
        raise click.ClickException(f"Error probing file {err.strip().splitlines()[-1]}") from e

    yield probed


@contextlib.contextmanager
def probe_audio_streams(file: Path) -> Iterator[list[AudioStream]]:
    """Probe the audio streams of a file."""
    try:
        logger.info("Probing file %r", file)
        probe = ffmpeg.probe(file)

        audio_streams = sorted(
            [stream for stream in probe["streams"] if "codec_type" in stream and stream["codec_type"] == "audio"],
            key=lambda stream: stream["index"],
        )

        if not audio_streams:
            logger.warning("No audio streams found")
            yield []
            return

        logger.info("Found %d audio streams", len(audio_streams))
        yield audio_streams

    except (ffmpeg.Error, ValueError) as e:
        logger.critical("%s: %s", type(e).__name__, e)
        error_console.print_exception()
        raise e



"""Application commandline interface."""

import rich_click as click

from BAET._config.logging import app_logger, configure_logging, create_logger
from BAET.cli.help_configuration import baet_config

from .commands import extract, probe

logger = create_logger()


@click.group("baet")
@baet_config(use_markdown=True)
@click.version_option(prog_name="BAET", package_name="BAET", message="%(prog)s v%(version)s")
@click.option("--logging", "-L", help="Run the application with logging.", count=True)
def cli(logging: int) -> None:
    """**Bulk Audio Extraction Tool (BAET)**

    This tool provides a simple way to extract audio from video files.
    - You can use --help on any command to get more information.
    """  # noqa: D400
    configure_logging(enable_logging=logging > 0)
    if logging > 1:
        app_logger.setLevel("DEBUG")

    # Currently only two levels of verbosity for info/debug level logging
    logger.info("Logging verbosity: %s", max(0, min(2, logging)))


cli.add_command(extract.extract)
cli.add_command(probe.probe)



"""Extract click command."""

import re
from collections.abc import Callable, MutableMapping, Sequence
from dataclasses import dataclass, field
from functools import wraps
from pathlib import Path
from re import Pattern
from typing import Concatenate

import rich.repr
import rich_click as click
from rich.console import Group
from rich.live import Live
from rich.padding import Padding
from rich.pretty import pretty_repr
from rich.table import Table

import ffmpeg
from BAET._config.console import app_console
from BAET._config.logging import create_logger
from BAET.cli.help_configuration import baet_config
from BAET.constants import AUDIO_EXTENSIONS, VIDEO_EXTENSIONS_NO_DOT, VideoExtension_NoDot
from BAET.display.job_progress import FFmpegJobProgress
from BAET.FFmpeg.jobs import AudioExtractJob
from BAET.FFmpeg.probe import probe_audio_streams
from BAET.helpers.string_helpers import pretty_join
from BAET.typing import AudioStream
from ffmpeg import Stream

logger = create_logger()


@rich.repr.auto()
@dataclass()
class ExtractJob:
    """Dataclass for holding extract job information."""

    input_outputs: list[tuple[Path, Path]] = field(default_factory=lambda: [])
    includes: list[Pattern[str]] = field(default_factory=lambda: [])
    excludes: list[Pattern[str]] = field(default_factory=lambda: [])
    include_extensions: list[Pattern[str]] = field(default_factory=lambda: [])


pass_extract_context = click.make_pass_decorator(ExtractJob, ensure=True)

type ExtractJobProcessor[**P] = Callable[P, Callable[[ExtractJob], ExtractJob]]


def processor[**P](
    f: Callable[Concatenate[ExtractJob, P], ExtractJob],
) -> ExtractJobProcessor[P]:
    """Produce an `ExtractJob`-accepting function from a function that accepts multiple arguments."""

    @wraps(f)
    def new_func(*args: P.args, **kwargs: P.kwargs) -> Callable[[ExtractJob], ExtractJob]:
        def _processor(job: ExtractJob) -> ExtractJob:
            updated_job: ExtractJob = f(job, *args, **kwargs)  # Need variable because MyPy doesn't understand PEP 695
            return updated_job

        return _processor

    return new_func


@click.group(chain=True, invoke_without_command=True)
@click.option(
    "--overwrite",
    is_flag=True,
    default=False,
    show_default=True,
    help="Overwrite existing files.",
)
@click.option(
    "--dry-run",
    is_flag=True,
    default=False,
    show_default=True,
    help="Run without actually producing any output.",
)
@baet_config()
def extract(dry_run: bool, overwrite: bool) -> None:
    """Extract click command."""


@extract.result_callback()
@click.pass_context
def process(
    ctx: click.Context,
    processors: Sequence[Callable[[ExtractJob], ExtractJob]],
    dry_run: bool,
    overwrite: bool,
) -> None:
    """Process the extract command."""
    logger.info("Dry run: %s", dry_run)

    job: ExtractJob = ExtractJob()
    for p in processors:
        job = p(job)

    if not job.include_extensions:
        default_filter = ctx.invoke(filter_command, extensions=[re.escape(e) for e in VIDEO_EXTENSIONS_NO_DOT])
        job = default_filter(job)

    logger.debug("Job (Prefiltered Inputs)::\n%s", pretty_repr(job))
    logger.info(pretty_join(job.input_outputs, "Prefiltered inputs", formatter=lambda inout: f"{inout[0]!r}"))

    for include in job.includes:
        job.input_outputs = list(filter(lambda x: include.match(x[0].name), job.input_outputs))

    for exclude in job.excludes:
        job.input_outputs = list(filter(lambda x: not exclude.match(x[0].name), job.input_outputs))

    job.input_outputs = list(
        filter(lambda x: any(ext.match(x[0].name) for ext in job.include_extensions), list(job.input_outputs))
    )

    logger.debug("Job (Filtered Inputs):\n%s", pretty_repr(job))
    logger.info(pretty_join(job.input_outputs, "Filtered inputs", formatter=lambda inout: f"{inout[0]!r}"))

    logger.info(
        pretty_join(
            job.input_outputs,
            "Extracting",
            formatter=lambda inout: f"{inout[0]!r} -> {inout[1].relative_to(inout[0].parent)!r}",
        )
    )

    built = [build_job(io[0], io[1]) for io in job.input_outputs]

    if not dry_run:
        run_synchronously(built)

    logger.info("Finished extracting.")


def build_job(file: Path, out_path: Path) -> AudioExtractJob:
    """Build an audio extraction job.

    Parameters
    ----------
    file : Path
        The file to extract audio from.

    out_path : Path
        The output path to extract to.

    Returns
    -------
    AudioExtractJob
        The audio extraction job.
    """
    out_path.parent.mkdir(parents=True, exist_ok=True)

    audio_streams: list[AudioStream] = []
    indexed_outputs: MutableMapping[int, Stream] = {}

    file = file.expanduser()
    with probe_audio_streams(file) as streams:
        for idx, stream in enumerate(streams):
            ffmpeg_input = ffmpeg.input(str(file))
            stream_index = stream["index"]
            output_path = out_path.with_stem(f"{out_path.stem}_track{stream_index}")
            sample_rate = stream.get(
                "sample_rate",
                44100,
            )

            audio_streams.append(stream)

            indexed_outputs[stream_index] = (
                ffmpeg.output(
                    ffmpeg_input[f"a:{idx}"],
                    f"{output_path.resolve().as_posix()}",  # .replace(" ", r"\ ")}",
                    format=output_path.suffix.lstrip("."),
                    acodec="pcm_s16le",
                    audio_bitrate=sample_rate,
                )
                .overwrite_output()
                .global_args("-progress", "-", "-nostats")
            )

    return AudioExtractJob(file, audio_streams, indexed_outputs)


def run_synchronously(jobs: list[AudioExtractJob]) -> None:
    """Run audio extraction jobs synchronously.

    Parameters
    ----------
    jobs : list[AudioExtractJob]
        The extraction jobs for FFmpeg to run.
    """
    display = Table.grid()

    job_progresses = [FFmpegJobProgress(job) for job in jobs]
    display.add_row(Padding(Group(*job_progresses), pad=(1, 2)))

    logger.info("Starting synchronous execution of queued jobs")
    with Live(display, console=app_console):
        for progress in job_progresses:
            logger.info("Starting job %r", {progress.job.input_file})
            progress.start()


@extract.command("file")
@click.option(
    "--input",
    "-i",
    "input_",
    help="The file to extract audio from.",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True, path_type=Path),
    required=True,
)
@click.option(
    "--output",
    "-o",
    help="The output file or directory to output to.",
    type=click.Path(exists=False, resolve_path=True, path_type=Path),
    default=None,
)
@click.option(
    "--filetype",
    "-f",
    help="The output filetype.",
    required=False,
    type=click.Choice(AUDIO_EXTENSIONS, case_sensitive=False),
    default=None,
)
@baet_config()
@processor
def input_file(job: ExtractJob, input_: Path, output: Path | None, filetype: str | None) -> ExtractJob:
    """Extract specific tracks from a video file."""
    if filetype is not None and not filetype.startswith("."):
        filetype = f".{filetype}"

    if output is None:
        out = input_.with_suffix(filetype or ".wav")
    elif output.is_file():
        if filetype:
            logger.warning("Provided a file output and filetype, ignoring filetype.")
        out = output
    elif output.is_dir():
        out = output / input_.with_suffix(filetype or ".wav").name
    else:
        raise click.BadParameter(f"Invalid output path: {output!r}", param_hint="output")

    logger.info("Extracting audio tracks from video file: %r", input_)
    logger.info("Extracting to: %r", out)

    job.input_outputs.append((input_, out))
    return job


@extract.command("dir")
@click.option(
    "--input",
    "-i",
    "input_",
    help="The directory of videos to extract audio from.",
    type=click.Path(exists=True, file_okay=False, resolve_path=True, path_type=Path),
    required=True,
)
@click.option(
    "--output",
    "-o",
    help="The output directory.",
    default=None,
    show_default="[INPUT]",
    type=click.Path(exists=False, file_okay=False, resolve_path=True, path_type=Path),
)
@click.option(
    "--filetype",
    "-f",
    help="The output filetype.",
    type=click.Choice(AUDIO_EXTENSIONS, case_sensitive=False),
    default="wav",
)
@baet_config()
@processor
def input_dir(job: ExtractJob, input_: Path, output: Path | None, filetype: str) -> ExtractJob:
    """Extract specific tracks from a video file."""
    if output is None:
        output = input_

    logger.info("Extracting audio tracks from video in dir: %r", input_)
    logger.info("Extracting to directory: %r", output)
    logger.info("Extracting to filetype: %r", filetype)

    if not filetype.startswith("."):
        filetype = f".{filetype}"

    for input_file in input_.iterdir():
        if not input_file.is_file():
            continue
        file_output = output / input_file.stem / input_file.with_suffix(filetype).name
        job.input_outputs.append((input_file, file_output))

    return job


@extract.command("filter")
@click.option(
    "--include",
    "includes",
    multiple=True,
    show_default=False,
    default=[],
    help="Include files matching this pattern.",
)
@click.option(
    "--exclude",
    "excludes",
    multiple=True,
    show_default=False,
    default=[],
    help="Exclude files matching this pattern.",
)
@click.option(
    "--ext",
    "extensions",
    help="Specify which video extensions to include in the directory.",
    multiple=True,
    type=click.Choice(VIDEO_EXTENSIONS_NO_DOT, case_sensitive=False),
    default=VIDEO_EXTENSIONS_NO_DOT,
)
@click.option(
    "--case-sensitive/--case-insensitive",
    "case_sensitive",
    default=False,
    help="Whether to match case sensitively.",
    show_default=True,
)
@baet_config()
@processor
def filter_command(
    job: ExtractJob,
    includes: Sequence[str],
    excludes: Sequence[str],
    extensions: Sequence[VideoExtension_NoDot],
    case_sensitive: bool,
) -> ExtractJob:
    """Filter files for selection when providing a directory."""
    flag = re.IGNORECASE if not case_sensitive else re.NOFLAG

    include_patterns: list[Pattern[str]] = []
    exclude_patterns: list[Pattern[str]] = []
    extensions_patterns: list[Pattern[str]] = []

    pattern_list_pairs = zip(
        [includes, excludes, extensions],
        [include_patterns, exclude_patterns, extensions_patterns],
        strict=True,
    )

    try:
        for patterns, lst in pattern_list_pairs:
            for pattern in patterns:
                lst.append(re.compile(pattern, flag))
    except re.error as e:
        logger.error("Error: %s", e)
        raise click.BadParameter("Invalid pattern", param_hint="pattern") from e

    include_patterns = [re.compile(p, flag) for p in includes]
    exclude_patterns = [re.compile(p, flag) for p in excludes]
    extensions_patterns = [re.compile(rf".*\.{re.escape(e)}$", flag) for e in extensions]

    logger.info("Filtering is case sensitivity: %s", case_sensitive)

    if includes:
        logger.info(pretty_join(includes, "Include file patterns"))

    if excludes:
        logger.info(pretty_join(excludes, "Exclude file patterns"))

    if extensions:
        logger.info(pretty_join(extensions, "Including extensions"))
        logger.info(pretty_join(extensions_patterns, "Including' extensions", formatter=lambda p: p.pattern))

    job.includes.extend(include_patterns)
    job.excludes.extend(exclude_patterns)
    job.include_extensions.extend(extensions_patterns)

    return job



"""Custom type to parse cli args with click."""

import re
from collections import OrderedDict
from collections.abc import Callable, Iterable
from re import Pattern
from typing import Any, Protocol, override, runtime_checkable

import rich_click as click
from rich.repr import Result


class RegexPatternParamType(click.ParamType):
    """Regex Pattern type for parsing with click."""

    name = "RegexPattern"

    @override
    def convert(self, value: Any, param: click.Parameter | None, ctx: click.Context | None) -> Pattern[str]:
        if isinstance(value, Pattern):
            return value

        pattern: Pattern[str]
        try:
            pattern = re.compile(value)
        except ValueError:
            self.fail(f"'{value!r}' is an invalid pattern...", param, ctx)

        return pattern


RegexPattern = RegexPatternParamType()


@runtime_checkable
class Equatable(Protocol):
    """A protocol asserting that an object has an implementation of `__eq__`."""

    def __eq__(self, __value: object) -> bool:  # noqa: D105
        ...


class Merger[T: Equatable, K: Equatable]:
    """A class to merge elements of a list, compared via a comparer and composed via a composer function.

    This class is effectively an aggregator or folder.
    """

    def __init__(self, selector: Callable[[T], K], composer: Callable[[T, T], T]) -> None:
        self._selector = selector
        self._composer = composer

    def __call__(self, it: Iterable[T]) -> list[T]:
        """Merge the elements of the given iterable.

        Elements are compared using the provided selector.
        Elements that are found to be equal are merged using the provided composer.

        Parameters
        ----------
        it : Iterable[T] The iterable of elements to merge.


        Returns
        -------
        list[T] A list of merged elements.
        """
        merged: OrderedDict[K, T] = OrderedDict()
        for item in it:
            key = self._selector(item)
            if key not in merged:
                merged[key] = item
                continue

            merged[key] = self._composer(merged[key], item)

        return list(merged.values())


class FFmpegArgsRepr:
    """A class to represent FFmpeg arguments for rich pretty-printy."""

    def __init__(self, args: list[str]) -> None:
        self.args = args

    def __rich_repr__(self) -> Result:
        """Rich representation of the object."""
        yield self.args


FFmpegArgsRepr.__name__ = "FFmpeg"



"""Help configuration for the CLI."""

from collections.abc import Callable
from functools import wraps
from typing import Any

import rich_click as click
from rich_click import RichHelpConfiguration
from rich_click.rich_help_configuration import OptionHighlighter

from BAET._config.console import app_console


def baet_config(
    show_arguments: bool = True,
    column_width_ratio: tuple[None, None] | tuple[int, int] = (None, None),
    use_markdown: bool = False,
    use_markdown_emoji: bool = True,
    use_rich_markup: bool = True,
) -> Any:
    """Apply a consistent help command as a function decorator."""

    def wrapper(func: Callable[..., Any]) -> Any:
        @click.rich_config(
            console=app_console,
            help_config=RichHelpConfiguration(
                show_arguments=show_arguments,
                style_commands_table_column_width_ratio=column_width_ratio,
                use_markdown=use_markdown,
                use_markdown_emoji=use_markdown_emoji,
                use_rich_markup=use_rich_markup,
                style_helptext="not dim bright_white",
                highlighter=OptionHighlighter(),
            ),
        )
        @wraps(func)
        def inner(*args: Any, **kwargs: Any) -> Any:
            return func(*args, **kwargs)

        return inner

    return wrapper



"""Functions for validating files."""

from pathlib import Path

from BAET.constants import VIDEO_EXTENSIONS


def file_is_supported_video(file: Path) -> bool:
    """Check if the file is supported a video file.

    Parameters
    ----------
    file : Path
        The file to check.

    Returns
    -------
    bool
        True if the file is a supported video file, False otherwise.
    """
    return file.is_file() and file.suffix in VIDEO_EXTENSIONS



"""Bulk Audio Export Tool (BAET) is a command line tool for exporting audio tracks from video files in bulk."""

import sys

from rich.traceback import install

from BAET.cli.cli import cli

# Enable rich traceback
install(show_locals=True)


def main() -> None:
    """Entry point for BAET."""
    cli()
    sys.exit(0)



"""Constants for BAET."""

import typing
from collections.abc import Sequence
from typing import Final, Literal

# TODO: Refactor required
VideoExtension = Literal[".mp4", ".mkv", ".avi", ".webm"]
VideoExtension_NoDot = Literal["mp4", "mkv", "avi", "webm"]

# TODO: Refactor required
VIDEO_EXTENSIONS: Final[Sequence[VideoExtension]] = [".mp4", ".mkv", ".avi", ".webm"]
VIDEO_EXTENSIONS_NO_DOT: Final[tuple[VideoExtension_NoDot, ...]] = typing.get_args(VideoExtension_NoDot)

AudioExtension = Literal["mp3", "wav", "flac", "ogg"]
AUDIO_EXTENSIONS: Final[tuple[AudioExtension, ...]] = typing.get_args(AudioExtension)



"""Bulk Audio Export Tool (BAET) is a command line tool for exporting audio tracks from video files in bulk."""

from importlib.metadata import version

from ._config.console import app_console
from ._config.logging import configure_logging, create_logger
from .theme import app_theme

__version__ = version(__name__)

__all__ = ["app_console", "configure_logging", "create_logger", "app_theme"]



"""Theme options used by Rich library for the app."""

from typing import Final

from rich.theme import Theme

app_theme_dict: Final[dict[str, str]] = {
    "app.version": "italic bright_cyan",
    # Help screen
    "keyword": "bold blue",
    "todo": "reverse bold indian_red",
    # Progress status
    "status.waiting": "dim bright_white",
    "status.running": "bright_cyan",
    "status.completed": "bright_green",
    "status.error": "bright_red",
}

app_theme = Theme(app_theme_dict)
